@@ -65,7 +65,10 @@ example, if our first expectation had been 3 instead of 4, we'd see:
 
 ```txt
 ✗ example.py:10
-  Result 4 was different from the expected value 3.
+  Result
+    4
+  was different from the expected value
+    3
 ```
 
 ## Core functionality
@@ -110,7 +113,7 @@ TODO: Prevent crash if expr_src is unavailable when tracing...
 
 # TODO: Cache compiled ASTs!
 
-__version__ = "0.14"
+__version__ = "0.16"
 
 import sys
 import inspect
@@ -119,6 +122,8 @@ import copy
 import io
 import types
 import builtins
+import cmath
+import textwrap
 
 
 #---------#
@@ -192,6 +197,18 @@ since trailing whitespace is hard to reason about because it's
 invisible.
 """
 
+FLOAT_REL_TOLERANCE = 1e-8
+"""
+The relative tolerance for floating-point similarity (see
+`cmath.isclose`).
+"""
+
+FLOAT_ABS_TOLERANCE = 1e-8
+"""
+The absolute tolerance for floating-point similarity (see
+`cmath.isclose`).
+"""
+
 
 #--------#
 # Errors #
@@ -478,7 +495,7 @@ def runFile():
     attribute is one of these names.
     """
     # Grab filename of the calling code from the stack
-    filename = get_my_location()["file"]
+    filename = get_my_location(speculate_filename=False)["file"]
 
     # Read the calling code's file
     try:
@@ -709,15 +726,17 @@ def expr_details(context):
     """
     # Expression that was evaluated
     expr = context.get("expr_src", "???")
-    short_expr = ellipsis(expr)
+    short_expr = ellipsis(expr, 78)
     # Results
     msg = ""
     extra_msg = ""
 
     # Base message
-    msg += f"In expression {short_expr}, values were:"
+    msg += f"Test expression was:\n{indent(short_expr, 2)}"
+
+    # Figure out values to display
     vdict = context.get("values", {})
-    if context.get("relevant"):
+    if context.get("relevant") is not None:
         show = sorted(
             context["relevant"],
             key=lambda fragment: (expr.index(fragment), len(fragment))
@@ -728,6 +747,9 @@ def expr_details(context):
             key=lambda fragment: (expr.index(fragment), len(fragment))
         )
 
+    if len(show) > 0:
+        msg += "\nValues were:"
+
     longs = []
     for key in show:
         if key in vdict:
@@ -754,7 +776,7 @@ def expr_details(context):
         ],
         key=lambda fragment: (expr.index(fragment), len(fragment))
     )
-    if context.get("relevant") and extra_values:
+    if context.get("relevant") is not None and extra_values:
         if extra_msg != "" and not extra_msg.endswith('\n'):
             extra_msg += '\n'
         extra_msg += "Extra values:"
@@ -765,7 +787,7 @@ def expr_details(context):
                 val = "???"
 
             entry = f"  {ev} = {val}"
-            fits = ellipsis(entry)
+            fits = ellipsis(entry, 78)
             extra_msg += '\n' + fits
             if fits != entry:
                 longs.append(entry)
@@ -802,8 +824,8 @@ def check_expectation(exp, set_success=True):
     result = case["result"]
     if exp["type"] == "result":
         expected = exp["value"]
-        short_result = ellipsis(repr(result))
-        short_expected = ellipsis(repr(expected))
+        short_result = ellipsis(repr(result), 78)
+        short_expected = ellipsis(repr(expected), 78)
         full_result = repr(result)
         full_expected = repr(expected)
         if checkEquality(result, expected):
@@ -811,8 +833,8 @@ def check_expectation(exp, set_success=True):
                 exp["tag"],
                 context,
                 (
-                    f"Result {short_result} was equivalent to the"
-                    f" expected value {short_expected}."
+                    f"Result:\n  {short_result}\nwas equivalent to the"
+                    f" expected value:\n  {short_expected}"
                 )
             )
             print_message(msg, color="34" if COLORS else None)
@@ -823,8 +845,8 @@ def check_expectation(exp, set_success=True):
         else:
             # Construct base report
             base_msg = (
-                f"Result {short_result} was different from the"
-                f" expected value {short_expected}."
+                f"Result:\n  {short_result}\nwas different from the"
+                f" expected value:\n  {short_expected}"
             )
 
             # Construct extra-detailed message
@@ -869,16 +891,22 @@ def check_expectation(exp, set_success=True):
                 exp["message"] = msg
             return False
         short_output = ellipsis(output)
+        if not short_output.endswith('\n'):
+            short_output += '\\\n'
         short_fragment = ellipsis(fragment)
         full_output = output
         full_fragment = fragment
+        if not full_output.endswith('\n'):
+            full_output += '\\\n'
+        if not full_fragment.endswith('\n'):
+            full_fragment += '\\\n'
         if checkContainment(fragment, output):
             msg = create_success_message(
                 exp["tag"],
                 context,
                 (
-                    f'Fragment "{short_fragment}" was found in the'
-                    f' recorded output """{short_output}""".'
+                    f'Fragment:\n  "{short_fragment}"\nwas found in the'
+                    f' recorded output:\n"""\\\n{short_output}"""'
                 )
             )
             print_message(msg, color='34' if COLORS else None)
@@ -889,22 +917,22 @@ def check_expectation(exp, set_success=True):
         else:
             # Construct base report
             msg = (
-                f'Fragment "{short_fragment}" was NOT present in the'
-                f' recorded output """{short_output}""".'
+                f'Fragment:\n  "{short_fragment}"\nwas NOT present in the'
+                f' recorded output:\n"""\\\n{short_output}"""'
             )
             if fragment.casefold() in output.casefold():
-                msg += "\n  Note: fragment would match if case were ignored."
+                msg += "\nNote: fragment would match if case were ignored."
 
             # Construct extra-detailed message
             extra_msg = ""
             if full_output != short_output:
                 extra_msg += (
-                    f'\nFull output:\n"""\n{full_output}\n"""'
+                    f'\nFull output:\n"""\\\n{full_output}\n"""'
                     f' ({len(full_output)} characters)'
                 )
             if full_fragment != short_fragment:
                 extra_msg += (
-                    f'\nFull fragment:\n"""\n{full_fragment}\n"""'
+                    f'\nFull fragment:\n"""\\\n{full_fragment}\n"""'
                     f' ({len(full_fragment)} characters)'
                 )
 
@@ -1069,6 +1097,10 @@ def expectCustom(tester):
         return check_expectation(exp)
 
 
+#------------#
+# Comparison #
+#------------#
+
 def checkEquality(val1, val2):
     """
     Returns True if val1 is 'equal' to val2, and False otherwise.
@@ -1076,13 +1108,13 @@ def checkEquality(val1, val2):
     whitespace in two strings when comparing them for equality.
     """
     if (not isinstance(val1, str)) or (not isinstance(val2, str)):
-        return val1 == val2 # use regular equality test
+        return compare(val1, val2) # use regular equality test
     # For two strings, pay attention to IGNORE_TRAILING_WHITESPACE
     elif IGNORE_TRAILING_WHITESPACE:
         # remove trailing whitespace from both strings (on all lines)
-        return trimWhitespace(val1) == trimWhitespace(val2)
+        return compare(trimWhitespace(val1), trimWhitespace(val2))
     else:
-        return val1 == val2 # use regular equality test
+        return compare(val1, val2) # use regular equality test
 
 
 def checkContainment(val1, val2):
@@ -1110,6 +1142,87 @@ def trimWhitespace(st):
     return '\n'.join(line.rstrip() for line in st.splitlines())
 
 
+def compare(val, ref, comparing=None):
+    """
+    Compares two values, allowing a bit of difference in terms of
+    floating point numbers, including numbers in complex structures.
+    Returns True if the two arguments are equivalent and false if not.
+
+    Works for recursive data structures.
+    """
+    if comparing is None:
+        comparing = set()
+
+    cmpkey = (id(val), id(ref))
+    if cmpkey in comparing:
+        # Either they differ somewhere else, or they're functionally
+        # identical
+        # TODO: Does this really ward off all infinite recursion on
+        # finite structures?
+        return True
+
+    comparing.add(cmpkey)
+
+    if val == ref:
+        return True
+    else: # let's hunt for differences
+        if (
+            isinstance(val, (int, float, complex))
+        and isinstance(ref, (int, float, complex))
+        ): # what if they're both numbers?
+            return cmath.isclose(
+                val,
+                ref,
+                rel_tol=FLOAT_REL_TOLERANCE,
+                abs_tol=FLOAT_ABS_TOLERANCE
+            )
+        elif type(val) != type(ref): # different types; not both numbers
+            return False
+        elif isinstance(val, (list, tuple)): # both lists or tuples
+            if len(val) != len(ref):
+                return False
+            else:
+                return all(
+                    compare(val[i], ref[i], comparing)
+                    for i in range(len(val))
+                )
+
+        elif isinstance(val, (set)): # both sets
+            if len(val) != len(ref):
+                return False
+            onlyVal = (val - ref)
+            onlyRef = (ref - val)
+            # TODO: Faster here, but still handle float imprecision?!?
+            return compare(sorted(onlyVal), sorted(onlyRef), comparing)
+
+        elif isinstance(val, dict): # both dicts
+            if len(val) != len(ref):
+                return False
+
+            vkeys = set(val.keys())
+            rkeys = set(val.keys())
+            onlyVal = sorted(vkeys - rkeys)
+            onlyRef = sorted(rkeys - vkeys)
+            both = vkeys & rkeys
+
+            for key_index in range(len(onlyVal)):
+                vk = onlyVal[key_index]
+                rk = onlyRef[key_index]
+                if not compare(vk, rk, comparing):
+                    return False
+
+                if not compare(val[vk], ref[rk], comparing):
+                    return False
+
+            return all(
+                compare(val[k], ref[k], comparing)
+                for k in both
+            )
+
+        else: # not sure what kind of thing this is...
+            return val == ref
+
+
 #-----------------------#
 # Configuration control #
 #-----------------------#
@@ -1514,7 +1627,9 @@ def get_expr_src(src, call_node):
 
     # If get_source_segment is available, use that
     if hasattr(ast, "get_source_segment"):
-        return ast.get_source_segment(src, arg_expr)
+        return textwrap.dedent(
+            ast.get_source_segment(src, arg_expr)
+        ).strip()
     else:
         # We're going to have to do this ourself: find the start of the
         # expression and state-machine to find a matching paren
@@ -1522,7 +1637,7 @@ def get_expr_src(src, call_node):
         open_paren = src.index('(', start)
         end = find_closing_item(src, open_paren, '()')
         # Note: can't be None because that would have been a SyntaxError
-        return src[open_paren + 1:end]
+        return textwrap.dedent(src[open_paren + 1:end]).strip()
 
 
 def get_ref_src(src, node):
@@ -1666,13 +1781,20 @@ def get_module(stack_frame):
     return sys.modules.get(other_name)
 
 
-def get_filename(stack_frame):
+def get_filename(stack_frame, speculate_filename=True):
     """
     Given a stack frame, returns the filename of the file in which the
     code which created that stack frame was defined. Returns None if
-    that information isn't available.
+    that information isn't available via a __file__ global, or if
+    speculate_filename is True (the default), uses the value of the
+    frame's f_code.co_filename, which may not always be a real file on
+    disk, or which is weird circumstances could be the name of a file on
+    disk which is *not* where the code came from.
     """
-    return stack_frame.f_globals.get("__file__", None)
+    filename = stack_frame.f_globals.get("__file__")
+    if filename is None and speculate_filename:
+        filename = stack_frame.f_code.co_filename
+    return filename
 
 
 def get_code_line(stack_frame):
@@ -1792,10 +1914,9 @@ def walk_ast_in_order(node):
         elif hasattr(ast, "NamedExpr") and isinstance(node, ast.NamedExpr):
             yield from walk_ast_in_order(node.value)
             yield from walk_ast_in_order(node.target)
-        # TODO: HERE
-        #elif isinstance(node, ast.BinOp):
-        #    yield from walk_ast_in_order(node.left)
-        #    yield from walk_ast_in_order(node.right)
+        elif isinstance(node, ast.BinOp):
+            yield from walk_ast_in_order(node.left)
+            yield from walk_ast_in_order(node.right)
         elif isinstance(node, ast.UnaryOp):
             yield from walk_ast_in_order(node.operand)
         elif isinstance(node, ast.Lambda):
@@ -1863,10 +1984,6 @@ def walk_ast_in_order(node):
             yield from walk_ast_in_order(node.context_expr)
             yield from walk_ast_in_order(node.optional_vars)
         # alias and typeignore have no executable members
-        # TODO: HERE
-        elif isinstance(node, ast.BinOp):
-            yield from walk_ast_in_order(node.left)
-            yield from walk_ast_in_order(node.right)
 
         # Finally, yield this node itself
         yield node
@@ -1888,23 +2005,24 @@ def find_call_nodes_on_line(node, frame, function, lineno):
     ast.unparse is available, the string will also attempt to match
     (exactly) against the unparsed call expression.
 
-    Only calls that start on the given line number will match.
+    Calls that start on the given line number will match, but if there
+    are no such calls, then a call on a preceding line whose expression
+    includes the target line will be looked for and may match.
 
     The return value will be a list of ast.Call nodes, and they will be
     ordered in the same order that those nodes would be executed when
     the line of code is executed.
     """
-    result = []
-    for child in walk_ast_in_order(node):
-        # only consider call nodes on the target line
-        if (
-            hasattr(child, "lineno")
-        and child.lineno == lineno
-        and isinstance(child, ast.Call)
-        ):
-            call_expr = child.func
-            if isinstance(function, str):
-                if (
+    def call_matches(call_node):
+        """
+        Locally-defined matching predicate.
+        """
+        nonlocal function
+        call_expr = call_node.func
+        return (
+            (
+                isinstance(function, str)
+            and (
                     (
                         isinstance(call_expr, ast.Name)
                     and call_expr.id == function
@@ -1917,12 +2035,68 @@ def find_call_nodes_on_line(node, frame, function, lineno):
                         hasattr(ast, "unparse")
                     and ast.unparse(call_expr) == function
                     )
-                ):
-                    result.append(child)
-            else:
-                callee = evaluate_in_context(call_expr, frame)
-                if callee is function:
-                    result.append(child)
+                )
+            )
+         or (
+                not isinstance(function, str)
+            and evaluate_in_context(call_expr, frame) is function
+            )
+        )
+
+    result = []
+    all_on_line = []
+    for child in walk_ast_in_order(node):
+        # only consider call nodes on the target line
+        if (
+            hasattr(child, "lineno")
+        and child.lineno == lineno
+        ):
+            all_on_line.append(child)
+            if isinstance(child, ast.Call) and call_matches(child):
+                result.append(child)
+
+    # If we didn't find any candidates, look outwards from ast nodes on
+    # the target line to find a Call that encompasses them...
+    if len(result) == 0:
+        for on_line in all_on_line:
+            here = getattr(on_line, "parent", None)
+            while (
+                here is not None
+            and not isinstance(
+                    here,
+                    # Call (what we're looking for) plus most nodes that
+                    # indicate there couldn't be a call grandparent:
+                    (
+                        ast.Call,
+                        ast.Module, ast.Interactive, ast.Expression,
+                        ast.FunctionDef, ast.AsyncFunctionDef,
+                        ast.ClassDef,
+                        ast.Return,
+                        ast.Delete,
+                        ast.Assign, ast.AugAssign, ast.AnnAssign,
+                        ast.For, ast.AsyncFor,
+                        ast.While,
+                        ast.If,
+                        ast.With, ast.AsyncWith,
+                        ast.Raise,
+                        ast.Try,
+                        ast.Assert,
+                        ast.Assert,
+                        ast.Assert,
+                        ast.Assert,
+                        ast.Assert,
+                        ast.Assert,
+                        ast.Assert,
+                        ast.Assert,
+                    )
+                )
+            ):
+                here = getattr(here, "parent", None)
+
+            # If we found a Call that includes the target line as one
+            # of its children...
+            if isinstance(here, ast.Call) and call_matches(here):
+                result.append(here)
 
     return result
 
@@ -1953,15 +2127,20 @@ def is_inside_call_func(node):
         return is_inside_call_func(node.parent)
 
 
-def get_my_location():
+def get_my_location(speculate_filename=True):
     """
     Fetches the filename and line number of the external module which
     calls into this module. Returns a dictionary with "file" and "line"
     keys.
+
+    If speculate_filename is False, then the filename will be set to
+    None in cases where a __file__ global cannot be found, instead of
+    using f_code.co_filename as a backup. In some cases, this is useful
+    because f_code.co_filename may not be a valid file.
     """
     frame = get_external_calling_frame()
     try:
-        filename = get_filename(frame)
+        filename = get_filename(frame, speculate_filename)
         lineno = get_code_line(frame)
     finally:
         del frame
@@ -2005,28 +2184,42 @@ def get_my_context(function_or_name):
     try:
         filename = get_filename(frame)
         lineno = get_code_line(frame)
-        try:
-            with open(filename, 'r') as fin:
-                src = fin.read()
-        except Exception:
-            print(
-                "Warning: Unable to retrieve calling line's source.",
-                file=sys.stderr
-            )
-            print(
-                "Call is on line {} of module {} from file '{}'".format(
-                    lineno,
-                    frame.f_globals.get("__name__"),
-                    filename
-                ),
-                file=sys.stderr
-            )
+        if filename is None:
+            src = None
+        else:
+            try:
+                with open(filename, 'r') as fin:
+                    src = fin.read()
+            except Exception:
+                # We'll assume here that the source is something like an
+                # interactive shell so we won't warn unless the detail
+                # level is turned up.
+                if DETAIL_LEVEL >= 2:
+                    print(
+                        "Warning: unable to get calling code's source.",
+                        file=sys.stderr
+                    )
+                    print(
+                        (
+                            "Call is on line {} of module {} from file"
+                            " '{}'"
+                        ).format(
+                            lineno,
+                            frame.f_globals.get("__name__"),
+                            filename
+                        ),
+                        file=sys.stderr
+                    )
+                src = None
+
+        if src is None:
             return {
                 "file": filename,
                 "line": lineno
             }
 
         src_node = ast.parse(src, filename=filename, mode='exec')
+        assign_parents(src_node)
         candidates = find_call_nodes_on_line(
             src_node,
             frame,
